#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define BUFFER_SIZE 5
#define NUMBER_OF_ITEMS 10

//-------------------------------------------
typedef struct {
    int buffer[BUFFER_SIZE];
    size_t count;
    pthread_mutex_t mutex;
    pthread_cond_t not_empty;
    pthread_cond_t not_full;
} Buffer;

//-------------------------------------------
void init_buffer(Buffer *buf) {
    buf->count = 0;
    pthread_mutex_init(&buf->mutex, NULL);
    pthread_cond_init(&buf->not_empty, NULL);
    pthread_cond_init(&buf->not_full, NULL);
}

//-------------------------------------------
void destroy_buffer(Buffer *buf) {
    pthread_mutex_destroy(&buf->mutex);
    pthread_cond_destroy(&buf->not_empty);
    pthread_cond_destroy(&buf->not_full);
}

//-------------------------------------------
void *producer(void *param) {
    Buffer *buf = (Buffer *)param;
    for (int i = 0; i < NUMBER_OF_ITEMS; ++i) {
        pthread_mutex_lock(&buf->mutex);
        while (buf->count == BUFFER_SIZE)
            pthread_cond_wait(&buf->not_full, &buf->mutex);
        
        buf->buffer[buf->count++] = i;
        printf("Produced: %d\n", i);
        
        pthread_cond_signal(&buf->not_empty);
        pthread_mutex_unlock(&buf->mutex);
        
        sleep(1);  // Simulate production time
    }
    return NULL;
}

//-------------------------------------------
void *consumer(void *param) {
    Buffer *buf = (Buffer *)param;
    for (int i = 0; i < NUMBER_OF_ITEMS; ++i) {
        pthread_mutex_lock(&buf->mutex);
        while (buf->count == 0)
            pthread_cond_wait(&buf->not_empty, &buf->mutex);
        
        int item = buf->buffer[--buf->count];
        printf("Consumed: %d\n", item);
        
        pthread_cond_signal(&buf->not_full);
        pthread_mutex_unlock(&buf->mutex);
        
        sleep(2);  // Simulate consumption time
    }
    return NULL;
}

//-------------------------------------------
int main() {
    pthread_t prod_thread, cons_thread;
    Buffer buffer;
    
    init_buffer(&buffer);
    
    
    pthread_create(&prod_thread, NULL, producer, &buffer);
    pthread_create(&cons_thread, NULL, consumer, &buffer);
    
    pthread_join(prod_thread, NULL);
    printf("producer thread terminated\n");
    pthread_join(cons_thread, NULL);
    printf("Consumer thread terminated\n");
   
    destroy_buffer(&buffer);
    
    return 0;
}
